@@ -73,12 +73,48 @@ def _get_action(statement):
     return action
 
 
+def _get_bucket_resource(bucket_name):
+    """
+    :param bucket_name: Name of the bucket
+    :type bucket_name: str
+
+    :return: Representation of the bucket with the resource prefix
+    :rtype: str
+    """
+    return [_AWS_RESOURCE_PREFIX + bucket_name]
+
+
+def _get_resource_prefix(bucket_name):
+    """
+    :param bucket_name: Name of the bucket
+    :type bucket_name: str
+
+    :return: Representation of the bucket with the resource prefix,
+             but with a trailing slash.
+    :rtype: str
+    """
+    return _get_bucket_resource(bucket_name)[0] + '/'
+
+
+def _get_object_resource(bucket_name, prefix):
+    """
+    :param bucket_name: Name of the bucket
+    :type bucket_name: str
+    :param prefix: Name of the prefix
+    :type prefix: str
+
+    :return: Representation of an object in a bucket with the resource prefix.
+    :rtype: str
+    """
+    return [_get_resource_prefix(bucket_name) + prefix + '*']
+
+
 # Returns new statements with bucket actions.
 def _new_bucket_statement(policy, bucket_name, prefix=''):
     if policy == Policy.NONE:
         return []
 
-    bucket_resource = _AWS_RESOURCE_PREFIX + bucket_name
+    bucket_resource = _get_bucket_resource(bucket_name)
 
     rv = [{'Action': list(_COMMON_BUCKET_ACTIONS),
            'Effect': 'Allow',
@@ -112,7 +148,7 @@ def _new_object_statement(policy, bucket_name, prefix=''):
     rv = [{'Action': [],
            'Effect': 'Allow',
            'Principal': {'AWS': '*'},
-           'Resource': _AWS_RESOURCE_PREFIX + bucket_name + '/' + prefix + '*',
+           'Resource': _get_object_resource(bucket_name, prefix),
            'Sid': ''}]
     if policy == Policy.READ_ONLY:
         rv[0]['Action'] = list(_READ_ONLY_OBJECT_ACTIONS)
@@ -141,8 +177,8 @@ def _filter_resources(prefix, resources):
 # Returns whether given bucket statements are used by other than given
 # prefix statements.
 def _get_in_use_policy(statements, bucket_name, prefix=''):
-    resource_prefix = _AWS_RESOURCE_PREFIX + bucket_name + '/'
-    object_resource = _AWS_RESOURCE_PREFIX + bucket_name + '/' + prefix + '*'
+    resource_prefix = _get_resource_prefix(bucket_name)
+    object_resource = _get_object_resource(bucket_name, prefix)
 
     in_use = {Policy.READ_ONLY: False,
               Policy.WRITE_ONLY: False}
@@ -250,14 +286,15 @@ def _remove_object_actions(statement, policy):
 # Returns statements containing removed actions/statements for given
 # policy, bucket name and prefix.
 def _remove_statements(statements, policy, bucket_name, prefix=''):
-    bucket_resource = _AWS_RESOURCE_PREFIX + bucket_name
-    object_resource = _AWS_RESOURCE_PREFIX + bucket_name + '/' + prefix + '*'
+    bucket_resource = _get_bucket_resource(bucket_name)
+    object_resource = _get_object_resource(bucket_name, prefix)
     in_use = _get_in_use_policy(statements, bucket_name, prefix)
     out = []
     read_only_bucket_statements = []
     s3_prefix_value = []
     for s in statements:
-        if s['Resource'] == bucket_resource:
+        resource = _get_resource(s)
+        if resource == bucket_resource:
             if s.get('Condition'):
                 s = _remove_bucket_actions(s, policy, prefix)
             else:
@@ -270,11 +307,11 @@ def _remove_statements(statements, policy, bucket_name, prefix=''):
                         policy == Policy.READ_WRITE and
                         not in_use[Policy.WRITE_ONLY]):
                     s = _remove_bucket_actions(s, Policy.WRITE_ONLY, prefix)
-        elif s['Resource'] == object_resource:
+        elif resource == object_resource:
             s = _remove_object_actions(s, policy)
 
         if s:
-            if (s['Resource'] == bucket_resource and
+            if (resource == bucket_resource and
                     set(_get_action(s)) & _READ_ONLY_BUCKET_ACTIONS and
                     s.get('Effect') == 'Allow' and
                     s.get('Principal') == {'AWS': '*'}):
@@ -284,7 +321,7 @@ def _remove_statements(statements, policy, bucket_name, prefix=''):
                 if isinstance(values, basestring):
                     values = [values]
 
-                s3_prefix_value += [(bucket_resource + '/' + v + '*')
+                s3_prefix_value += [(bucket_resource[0] + '/' + v + '*')
                                     for v in values]
 
                 if s3_prefix_value or not cond_value:
@@ -294,9 +331,9 @@ def _remove_statements(statements, policy, bucket_name, prefix=''):
             out.append(s)
 
     skip_bucket_statement = True
-    resource_prefix = _AWS_RESOURCE_PREFIX + bucket_name + '/'
+    resource_prefix = _get_resource_prefix(bucket_name)
     for s in out:
-        resource = s['Resource']
+        resource = _get_resource(s)
         if (_filter_resources(resource_prefix, resource) and
                 resource not in s3_prefix_value):
             skip_bucket_statement = False
@@ -388,7 +425,7 @@ def _append_statement(statements, statement):
 
 
 # Appends two statement lists.
-def append_statements(statements, append_statements):
+def _append_statements(statements, append_statements):
     for s in append_statements:
         _append_statement(statements, s)
 
@@ -465,10 +502,35 @@ def _get_object_policy(statement):
     return read_only, write_only
 
 
+def _get_permissions(s, resource, object_resource, matched_resource,
+                     bucket_resource, prefix, bucket_common_found,
+                     bucket_read_only, bucket_write_only):
+
+    obj_read_only, obj_write_only = False, False
+    if (resource == object_resource or
+            fnmatch.fnmatch(object_resource, resource)):
+        read_only, write_only = _get_object_policy(s)
+        if len(matched_resource) < len(resource):
+            obj_read_only = read_only
+            obj_write_only = write_only
+            matched_resource = resource
+        elif len(matched_resource) == len(resource):
+            obj_read_only = obj_read_only or read_only
+            obj_write_only = obj_write_only or write_only
+            matched_resource = resource
+    elif resource == bucket_resource:
+        common_found, read_only, write_only = _get_bucket_policy(s, prefix)
+        bucket_common_found = bucket_common_found or common_found
+        bucket_read_only = bucket_read_only or read_only
+        bucket_write_only = bucket_write_only or write_only
+    return bucket_common_found, bucket_read_only, bucket_write_only, \
+        obj_read_only, obj_write_only
+
+
 # Returns policy of given bucket name, prefix in given statements.
 def get_policy(statements, bucket_name, prefix=''):
-    bucket_resource = _AWS_RESOURCE_PREFIX + bucket_name
-    object_resource = _AWS_RESOURCE_PREFIX + bucket_name + '/' + prefix + '*'
+    bucket_resource = _get_bucket_resource(bucket_name)[0]
+    object_resource = _get_object_resource(bucket_name, prefix)[0]
 
     bucket_common_found = False
     bucket_read_only = False
@@ -477,23 +539,13 @@ def get_policy(statements, bucket_name, prefix=''):
     obj_read_only = False
     obj_write_only = False
     for s in statements:
-        resource = s['Resource']
-        if (resource == object_resource or
-                fnmatch.fnmatch(object_resource, resource)):
-            read_only, write_only = _get_object_policy(s)
-            if len(matched_resource) < len(resource):
-                obj_read_only = read_only
-                obj_write_only = write_only
-                matched_resource = resource
-            elif len(matched_resource) == len(resource):
-                obj_read_only = obj_read_only or read_only
-                obj_write_only = obj_write_only or write_only
-                matched_resource = resource
-        elif resource == bucket_resource:
-            common_found, read_only, write_only = _get_bucket_policy(s, prefix)
-            bucket_common_found = bucket_common_found or common_found
-            bucket_read_only = bucket_read_only or read_only
-            bucket_write_only = bucket_write_only or write_only
+        resources = _get_resource(s)
+        for resource in resources:
+            bucket_common_found, bucket_read_only, bucket_write_only, \
+                obj_read_only, obj_write_only = _get_permissions(
+                    s, resource, object_resource, matched_resource,
+                    bucket_resource, prefix, bucket_common_found,
+                    bucket_read_only, bucket_write_only)
 
     policy = Policy.NONE
     if bucket_common_found:
@@ -514,6 +566,6 @@ def set_policy(statements, policy, bucket_name, prefix=''):
     out = _remove_statements(statements, Policy.READ_WRITE,
                              bucket_name, prefix)
     ns = _new_statements(policy, bucket_name, prefix)
-    append_statements(out, ns)
+    _append_statements(out, ns)
 
     return out
